package com.qmth.exam.reserve.service.impl;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.qmth.boot.core.collection.PageResult;
import com.qmth.boot.core.concurrent.service.ConcurrentService;
import com.qmth.boot.core.exception.StatusException;
import com.qmth.boot.tools.excel.ExcelReader;
import com.qmth.boot.tools.excel.enums.ExcelType;
import com.qmth.boot.tools.excel.model.DataMap;
import com.qmth.boot.tools.io.ZipWriter;
import com.qmth.exam.reserve.bean.Constants;
import com.qmth.exam.reserve.bean.applytask.CurrentApplyTaskVO;
import com.qmth.exam.reserve.bean.category.CategoryCacheBean;
import com.qmth.exam.reserve.bean.login.LoginUser;
import com.qmth.exam.reserve.bean.stdapply.*;
import com.qmth.exam.reserve.cache.CacheConstants;
import com.qmth.exam.reserve.cache.impl.ApplyTaskCacheService;
import com.qmth.exam.reserve.cache.impl.CategoryCacheService;
import com.qmth.exam.reserve.cache.impl.ExamSiteCacheService;
import com.qmth.exam.reserve.dao.StudentApplyDao;
import com.qmth.exam.reserve.entity.*;
import com.qmth.exam.reserve.entity.base.BaseEntity;
import com.qmth.exam.reserve.enums.AsyncTaskType;
import com.qmth.exam.reserve.enums.CategoryLevel;
import com.qmth.exam.reserve.enums.EventType;
import com.qmth.exam.reserve.enums.Role;
import com.qmth.exam.reserve.service.*;
import com.qmth.exam.reserve.template.execute.StudentApplyNoFinishExportService;
import com.qmth.exam.reserve.util.DateUtil;
import com.qmth.exam.reserve.util.JsonHelper;
import com.qmth.exam.reserve.util.PageUtil;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.time.DateUtils;
import org.redisson.api.RLock;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.time.LocalDate;
import java.time.ZoneId;
import java.util.*;
import java.util.stream.Collectors;

import static org.apache.commons.lang3.time.DateUtils.isSameDay;

/**
 * @Description
 * @Author haoguanghui
 * @date 2024/06/24
 */
@Service
public class StudentApplyServiceImpl extends ServiceImpl<StudentApplyDao, StudentApplyEntity> implements StudentApplyService {

    private static final Logger log = LoggerFactory.getLogger(StudentApplyServiceImpl.class);

    private static final String[] EXCEL_HEADER = new String[] { "学号", "姓名", "证件号", "所属教学点", "预约考点1", "预约时段1", "预约考点2", "预约时段2", "预约考点3",
            "预约时段3", "预约考点4", "预约时段4" };

    @Autowired
    private ApplyTaskService applyTaskService;

    @Autowired
    private ExamSiteService examSiteService;

    @Autowired
    private TimePeriodService timePeriodService;

    @Autowired
    private ConcurrentService concurrentService;

    @Autowired
    private ApplyTaskCacheService cacheService;

    @Autowired
    private OperateLogService operateLogService;

    @Autowired
    private CategoryService categoryService;

    @Autowired
    private StudentService studentService;

    @Autowired
    private ExamSiteCacheService examSiteCacheService;

    @Autowired
    private CategoryCacheService categoryCacheService;

    @Autowired
    private StudentAutoAssignService studentAutoAssignService;

    @Autowired
    private MaterialGenerateService materialGenerateService;

    @Autowired
    private ExamRoomService examRoomService;

    @Autowired
    private AsyncTaskService asyncTaskService;

    @Autowired
    private StudentApplyNoFinishExportService studentApplyNoFinishExportService;


    @Override
    public PageResult<StudentApplyVO> page(StudentApplyReq req) {
        if (req.getTaskId() == null) {
            ApplyTaskEntity task = getApplyTask();
            req.setTaskId(task.getId());
        }
        if (req.getTeachingId() != null) {
            List<CategoryVO> examSiteList = examSiteService.listExamSite(req.getTeachingId(), Boolean.FALSE);
            if (examSiteList == null || examSiteList.isEmpty()) {
                return new PageResult<>();
            }
            List<Long> examSiteIds = examSiteList.stream().map(CategoryVO::getId).collect(Collectors.toList());
            req.setExamSiteIds(examSiteIds.isEmpty() ? null : examSiteIds);
        }
        IPage<StudentApplyVO> iPage = baseMapper.page(new Page<>(req.getPageNumber(), req.getPageSize()), req);
        List<StudentApplyVO> records = iPage.getRecords();
        for (StudentApplyVO record : records) {
            AgentAndRoomVO agentRoom = baseMapper.getAgentAndRoom(record.getId());
            record.setAgentName(agentRoom.getAgentName());
            record.setTeachingName(agentRoom.getTeachingName());
            record.setApplyTeachingName(agentRoom.getApplyTeachingName());
            record.setRoomName(agentRoom.getRoomName());
            record.setUserName(agentRoom.getUserName());
        }
        return PageUtil.of(iPage);
    }

    private ApplyTaskEntity getApplyTask() {
        LambdaQueryWrapper<ApplyTaskEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(ApplyTaskEntity::getEnable, Boolean.TRUE);
        ApplyTaskEntity task = applyTaskService.getOne(wrapper);
        if (task == null) {
            throw new StatusException("未开启预约任务");
        }
        return task;
    }

    @Transactional
    @Override
    public void cancel(LoginUser user, Long id) {
        StudentApplyEntity studentApplyEntity = getById(id);
        if (studentApplyEntity == null) {
            log.warn("[cancel] 考生预约信息不存在, id:{} userAccount:{}", id, user.getAccount());
            throw new StatusException("没有预约信息");
        }
        if (studentApplyEntity.getTimePeriodId() == null) {
            log.warn("[cancel] 考生预约的时段为空, id:{} userAccount:{}", id, user.getAccount());
            throw new StatusException("没有预约信息");
        }
        TimePeriodEntity timePeriod = timePeriodService.getById(studentApplyEntity.getTimePeriodId());
        if (timePeriod == null) {
            log.warn("[cancel] 考生预约的时段不存在, timePeriodId:{} userAccount:{}", studentApplyEntity.getTimePeriodId(), user.getAccount());
            throw new StatusException("考试时段不存在");
        }
        ApplyTaskEntity task = getApplyTask();
        Date applyDate = DateUtils.truncate(new Date(timePeriod.getStartTime()), Calendar.DATE);
        Date canCancelDay = DateUtil.addValues(applyDate, Calendar.DAY_OF_MONTH, -task.getAllowApplyCancelDays());
        if (new Date().after(canCancelDay)) {
            throw new StatusException("可取消时间已过，无法在取消");
        }

        String studentApplyLockKey = String.format(CacheConstants.LOCK_STUDENT_APPLY, studentApplyEntity.getStudentId());
        RLock studentApplyLock = (RLock) concurrentService.getLock(studentApplyLockKey);
        try {
            if (!studentApplyLock.tryLock()) {
                log.warn("[cancel] 获取锁失败，同一个考生不允许同时操作取消预约, lockKey:{}", studentApplyLockKey);
                throw new StatusException(Constants.SYSTEM_BUSY);
            } else {
                //先更新数据库
                studentApplyEntity.setCancel(Boolean.TRUE);
                updateById(studentApplyEntity);

                //扣减redis中的数量
                cacheService.decreaseApplyFinishCount(studentApplyEntity.getExamSiteId(), studentApplyEntity.getTimePeriodId());
            }
        } catch (Exception e) {
            log.error("[cancel] 取消预约失败, msg:{}", e.getMessage());
            throw new StatusException("取消预约失败，请稍后再试");
        } finally {
            try {
                if (studentApplyLock.isLocked() && studentApplyLock.isHeldByCurrentThread()) {
                    studentApplyLock.unlock();
                    log.info("[cancel] 解锁成功,lockKey:{}", studentApplyLock);
                }
            } catch (Exception e) {
                log.warn(e.getMessage());
            }
        }
        //写入日志
        operateLogService.insertOperateLog(user.getId(), EventType.CANCEL_APPLY, JsonHelper.toJson(studentApplyEntity));
    }

    @Transactional
    @Override
    public List<Map<String, Object>> importPreExam(LoginUser user, Long teachingId, Integer level, InputStream inputStream) {
        checkOpenTime(user.getOrgId());
        List<DataMap> lineList;
        ExcelReader reader = ExcelReader.create(ExcelType.XLSX, inputStream, 0);
        try {
            lineList = reader.getDataMapList();
        } catch (Exception e) {
            log.warn("[importPreExam] 读取预考数据失败,msg:{}", e.getMessage());
            throw new StatusException("Excel解析失败,请检查文件格式");
        }
        if (!Arrays.equals(EXCEL_HEADER, reader.getColumnNames())) {
            throw new StatusException("Excel表头错误,请不要修改模版中的表头");
        }
        if (CollectionUtils.isEmpty(lineList)) {
            throw new StatusException("Excel中没有任何考生预约数据");
        }
        //TODO 结果:如果这里加全局锁,执行后考生和其他教学点将无法再次执行;如果不加锁,这里影响大不大,待讨论
        RLock lock = (RLock) concurrentService.getLock(CacheConstants.LOCK_AUTO_APPLY);
        List<Map<String, Object>> failRecords;
        try {
            if (!lock.tryLock()) {
                log.warn("[importPreExam] 其他老师正在导入预考，获取锁失败！lockKey:{}", CacheConstants.LOCK_AUTO_APPLY);
                throw new StatusException("系统繁忙,请稍后重试！");
            }
            log.warn("[importPreExam] 获取锁成功！lockKey:{}", CacheConstants.LOCK_AUTO_APPLY);

            ApplyTaskEntity task = getApplyTask();
            Date openStartTime = new Date(task.getOpenApplyStartTime());
            Date now = new Date();
            failRecords = new ArrayList<>();
            List<StudentImportVO> applyList = new ArrayList<>();
            //教学点缓存
            Map<String, Long> teachingCache = getTeachingCache(level);
            //考点缓存
            Map<String, Long> agentCache = getAgentCache();
            //考试时段缓存
            Map<String, TimePeriodEntity> timePeriodCache = getTimePeriodCache();
            //考点所属教学点缓存
            Map<Long, Long> examSiteCategoryCache = getExamSiteCategoryCache();
            for (int i = 0; i < lineList.size(); i++) {
                DataMap line = lineList.get(i);
                List<AgentAndTimeVO> agentTimeList = new ArrayList<>();
                AgentAndTimeVO agentTime = new AgentAndTimeVO();
                StudentImportVO apply = new StudentImportVO();
                StringBuilder msg = new StringBuilder();

                String studentCode = trimAndNullIfBlank(line.get(EXCEL_HEADER[0]));
                if (StringUtils.isBlank(studentCode)) {
                    msg.append(" 学号不能为空");
                }

                String name = trimAndNullIfBlank(line.get(EXCEL_HEADER[1]));
                if (StringUtils.isBlank(name)) {
                    msg.append(" 姓名不能为空");
                }

                String identityNumber = trimAndNullIfBlank(line.get(EXCEL_HEADER[2]));
                if (StringUtils.isBlank(identityNumber)) {
                    msg.append(" 证件号码不能为空");
                }

                //判断考生是否存在
                StudentEntity studentEntity = checkStudent(studentCode, name, identityNumber, task);
                if (studentEntity == null) {
                    msg.append(" 考生信息填写错误");
                    failRecords.add(newError(i + 1, msg.toString()));
                    continue;
                } else {
                    apply.setStudentId(studentEntity.getId());
                    apply.setApplyNumber(studentEntity.getApplyNumber());
                }

                String teachingName = trimAndNullIfBlank(line.get(EXCEL_HEADER[3]));
                if (StringUtils.isBlank(teachingName)) {
                    msg.append(" 考生所属教学点不能为空");
                }
                Long categoryId = teachingCache.get(teachingName);
                if (categoryId == null) {
                    msg.append(" 考生所属教学点不存在");
                    failRecords.add(newError(i + 1, msg.toString()));
                    continue;
                }
                if (!studentEntity.getCategoryId().equals(categoryId)) {
                    msg.append(" 导入考生的所属教学点和系统中的考生教学点不匹配");
                }
                if (user.getRole().equals(Role.TEACHING) && categoryId.equals(teachingId)) {
                    msg.append(" 非本教学点考生");
                }

                String agentName1 = trimAndNullIfBlank(line.get(EXCEL_HEADER[4]));
                if (StringUtils.isBlank(agentName1)) {
                    msg.append(" 预约考点1不能为空");
                }
                Long applyCategoryId;
                Long agentId = agentCache.get(agentName1);
                if (agentId == null) {
                    msg.append(" 预约考点1不存在");
                } else {
                    applyCategoryId = examSiteCategoryCache.get(agentId);
                    if (now.before(openStartTime) && !categoryId.equals(applyCategoryId)) {
                        msg.append(" 未到自由预约时间,不能预约其他教学点下的考点");
                    }
                }

                String timePeriod1 = trimAndNullIfBlank(line.get(EXCEL_HEADER[5]));
                if (StringUtils.isBlank(timePeriod1)) {
                    msg.append(" 预约时段1不能为空");
                }
                TimePeriodEntity timePeriod;
                try {
                    timePeriod = checkTimePeriod(timePeriod1, timePeriodCache);
                    agentTime.setAgentId(agentId);
                    agentTime.setTimePeriodId(timePeriod.getId());
                    agentTime.setStartTime(timePeriod.getStartTime());
                    agentTimeList.add(agentTime);
                } catch (StatusException e) {
                    msg.append(" ").append(e.getMessage());
                }

                int index = 0;
                for (int num = 1; num <= 3; num++) {
                    String agentName = trimAndNullIfBlank(line.get(EXCEL_HEADER[num + 5 + index]));
                    String timePeriodName = trimAndNullIfBlank(line.get(EXCEL_HEADER[num + 6 + index]));
                    if (StringUtils.isBlank(agentName) && StringUtils.isBlank(timePeriodName)) {
                        apply.setAgentTimeList(agentTimeList);
                        applyList.add(apply);
                        if (msg.length() > 0)
                            failRecords.add(newError(i + 1, msg.toString()));
                        break;
                    } else {
                        agentId = agentCache.get(agentName);
                        if (agentId == null) {
                            msg.append(" 预约考点").append(num + 1).append("不存在");
                        } else {
                            applyCategoryId = examSiteCategoryCache.get(agentId);
                            if (now.before(openStartTime) && !applyCategoryId.equals(categoryId)) {
                                msg.append(" 未到自由预约时间，不允许预约其他教学点的考点");
                            }
                        }
                        try {
                            timePeriod = checkTimePeriod(timePeriodName, timePeriodCache);
                            agentTime = new AgentAndTimeVO();
                            agentTime.setAgentId(agentId);
                            agentTime.setTimePeriodId(timePeriod.getId());
                            agentTime.setStartTime(timePeriod.getStartTime());
                            agentTimeList.add(agentTime);
                        } catch (StatusException e) {
                            msg.append(" ").append(e.getMessage());
                        }
                    }

                    ++index;
                }
            }
            //检查导入excel的考生时段
            checkStudentApplyTime(applyList, task.getAllowApplyCancelDays(), failRecords);
            //过滤考生的可用时段
            filterAvailableTimePeriod(applyList, failRecords);

            if (CollectionUtils.isNotEmpty(failRecords)) {
                return failRecords;
            }

            //保存数据
            for (int i = 0; i < applyList.size(); i++) {
                StudentImportVO vo = applyList.get(i);
                try {
                    saveStudentApply(vo, user.getId());
                } catch (StatusException e) {
                    failRecords.add(newError(i + 1, "系统异常"));
                    log.error("[importPreExam] 导入异常", e);

                }
            }
        } catch (StatusException e) {
            log.error("[importPreExam] 导入预考失败,msg:{}", e.getMessage());
            throw new StatusException(e.getMessage());
        } finally {
            try {
                // 解锁前检查当前线程是否持有该锁
                if (lock.isLocked() && lock.isHeldByCurrentThread()) {
                    lock.unlock();
                    log.info("解锁成功！lockKey:{}", CacheConstants.LOCK_AUTO_APPLY);
                }
            } catch (Exception e) {
                log.warn(e.getMessage());
            }
        }
        return failRecords;
    }

    private void saveStudentApply(StudentImportVO vo, Long userId) {
        List<AgentAndTimeVO> agentTimeList = vo.getAgentTimeList();
        String studentApplyLockKey = String.format(CacheConstants.LOCK_STUDENT_APPLY, vo.getStudentId());
        RLock studentApplyLock = (RLock) concurrentService.getLock(studentApplyLockKey);
        try {
            if (!studentApplyLock.tryLock()) {
                log.warn("[importPreExam] 获取锁失败,考生在同时操作预约,lockKey:{}", studentApplyLockKey);
            } else {
                log.warn("[importPreExam] 获取锁成功,lockKey:{}", studentApplyLockKey);
                for (AgentAndTimeVO agentTimeVO : agentTimeList) {
                    StudentApplyEntity studentApply = new StudentApplyEntity();
                    studentApply.setStudentId(vo.getStudentId());
                    studentApply.setExamSiteId(agentTimeVO.getAgentId());
                    studentApply.setTimePeriodId(agentTimeVO.getTimePeriodId());
                    studentApply.setCancel(Boolean.FALSE);
                    studentApply.setOperateId(userId);
                    StudentApplyEntity existStudentApply = findStudentApply(studentApply);
                    if (existStudentApply != null) {
                        existStudentApply.setCancel(Boolean.FALSE);
                        baseMapper.updateById(existStudentApply);
                    } else {
                        baseMapper.insert(studentApply);
                    }
                    //redis的数量加1
                    cacheService.increaseApplyFinishCount(agentTimeVO.getAgentId(), agentTimeVO.getTimePeriodId());
                }
            }

        } catch (Exception e) {
            log.error("[importPreExam] 导入预约失败,msg:{}", e.getMessage());
            throw new StatusException("导入预考失败,请稍后重试");
        } finally {
            try {
                if (studentApplyLock.isLocked() && studentApplyLock.isHeldByCurrentThread()) {
                    studentApplyLock.unlock();
                    log.info("[importPreExam] 解锁成功！lockKey:{}", studentApplyLockKey);
                }
            } catch (Exception e) {
                log.warn(e.getMessage());
            }
        }
    }

    private StudentApplyEntity findStudentApply(StudentApplyEntity studentApply) {
        LambdaQueryWrapper<StudentApplyEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(StudentApplyEntity::getExamSiteId, studentApply.getExamSiteId());
        wrapper.eq(StudentApplyEntity::getTimePeriodId, studentApply.getTimePeriodId());
        wrapper.eq(StudentApplyEntity::getStudentId, studentApply.getStudentId());
        return baseMapper.selectOne(wrapper);
    }

    private void filterAvailableTimePeriod(List<StudentImportVO> applyList, List<Map<String, Object>> failRecords) {
        int index = 0;
        for (StudentImportVO apply : applyList) {
            index++;
            List<AgentAndTimeVO> agentTimeList = apply.getAgentTimeList();
            List<StudentApplyEntity> haveApplyList = listStudentHaveApply(apply.getStudentId());
            int studentApplyFinishCount = haveApplyList.size();
            agentTimeList.sort(Comparator.comparing(AgentAndTimeVO::getStartTime));

            List<AgentAndTimeVO> toBeApplyTimeList = new ArrayList<>();

            //考生已完成预约
            if (studentApplyFinishCount >= apply.getApplyNumber()) {
                apply.setAgentTimeList(toBeApplyTimeList);
                failRecords.add(newError(index, "请先取消考生已预约的信息"));
                continue;
            }

            //考生未完成预约
            List<AgentAndTimeVO> availableList = listAvailableTime(haveApplyList, agentTimeList);
            toBeApplyTimeList.addAll(availableList);

            //判断是否超过可预约次数
            if (apply.getApplyNumber() < haveApplyList.size() + availableList.size()) {
                failRecords.add(newError(index, "超过考生可预约的最大数量:" + apply.getApplyNumber()));
            } else {
                apply.setAgentTimeList(toBeApplyTimeList);
            }
        }
        //是否超过-考点的容量
        checkExamSiteTimeCapacity(applyList, failRecords);
    }

    private void checkExamSiteTimeCapacity(List<StudentImportVO> applyList, List<Map<String, Object>> failRecords) {
        try {
            List<AgentAndTimeVO> availableTimeList = new ArrayList<>();
            applyList.forEach(item -> availableTimeList.addAll(item.getAgentTimeList()));
            Map<Long, Map<Long, List<AgentAndTimeVO>>> agentTimeMap = availableTimeList.stream()
                    .collect(Collectors.groupingBy(AgentAndTimeVO::getAgentId, Collectors.groupingBy(AgentAndTimeVO::getTimePeriodId)));
            for (Map.Entry<Long, Map<Long, List<AgentAndTimeVO>>> examSiteEntry : agentTimeMap.entrySet()) {
                Long examSiteId = examSiteEntry.getKey();
                Map<Long, List<AgentAndTimeVO>> timePeriodMap = examSiteEntry.getValue();
                //考点的容量
                int examSiteCapacity = cacheService.getApplyTotalCount(examSiteId);
                for (Map.Entry<Long, List<AgentAndTimeVO>> timePeriodEntry : timePeriodMap.entrySet()) {
                    Long timePeriodKey = timePeriodEntry.getKey();
                    //考点下某个时段已预约的容量
                    int examSiteTimeCapacity = cacheService.getApplyFinishCount(examSiteId, timePeriodKey);
                    // 导入的某个考点，某个时段的数量
                    int impTimePeriodNum = timePeriodEntry.getValue().size();
                    if (examSiteCapacity < impTimePeriodNum + examSiteTimeCapacity) {
                        String examSiteName = examSiteCacheService.getExamSiteById(examSiteId).getExamSiteName();
                        TimePeriodEntity timePeriod = timePeriodService.getById(timePeriodKey);
                        String startAndEndTime = DateUtil.getStartAndEndTime(timePeriod.getStartTime(), timePeriod.getEndTime());
                        log.warn("[importPreExam],{} {} 剩余容量为:{},而导入的数量为:{}", examSiteName, startAndEndTime, examSiteCapacity - examSiteTimeCapacity, impTimePeriodNum);
                        failRecords.add(newError(0,
                                examSiteName + " " + startAndEndTime + " 剩余容量为:" + (examSiteCapacity - examSiteTimeCapacity) + ",而导入的数量为:" + impTimePeriodNum));
                    }
                }
            }
        } catch (Exception e) {
            log.error("[importPreExam],导入出错,msg:{}", e.getMessage());
            failRecords.add(newError(0, "导入失败"));
        }
    }

    private List<AgentAndTimeVO> listAvailableTime(List<StudentApplyEntity> haveApplyList, List<AgentAndTimeVO> agentTimeList) {
        List<AgentAndTimeVO> availableList = new ArrayList<>();
        boolean flag;
        for (AgentAndTimeVO agentTime : agentTimeList) {
            flag = false;
            for (StudentApplyEntity apply : haveApplyList) {
                if (agentTime.getTimePeriodId().equals(apply.getTimePeriodId())) {
                    flag = true;
                    break;
                }
            }
            if (!flag) {
                availableList.add(agentTime);
            }
        }
        return availableList;
    }

    private List<StudentApplyEntity> listStudentHaveApply(Long studentId) {
        LambdaQueryWrapper<StudentApplyEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(StudentApplyEntity::getStudentId, studentId);
        wrapper.eq(StudentApplyEntity::getCancel, Boolean.FALSE);
        return this.baseMapper.selectList(wrapper);
    }

    private void checkStudentApplyTime(List<StudentImportVO> applyList, Integer allowCancelDays, List<Map<String, Object>> failRecords) {
        for (int i = 0; i < applyList.size(); i++) {
            StudentImportVO vo = applyList.get(i);
            List<AgentAndTimeVO> agentTimeList = vo.getAgentTimeList();
            if (agentTimeList.size() > vo.getApplyNumber()) {
                failRecords.add(newError(i + 1, "考生的预约数量大于考生可预约的数量:" + vo.getApplyNumber()));
            }
            //是否重复预约时段
            boolean isDuplicate = agentTimeList.stream().collect(Collectors.groupingBy(AgentAndTimeVO::getTimePeriodId, Collectors.counting())).entrySet().stream()
                    .anyMatch(entry -> entry.getValue() > 1);
            if (isDuplicate) {
                failRecords.add(newError(i + 1, "同一个考生预约的时间不能相同"));
            }
            //预约的日期，是否在可取消日期之后
            LocalDate day = LocalDate.now().plusDays(allowCancelDays);
            long epochMilli = day.plusDays(1).atStartOfDay(ZoneId.systemDefault()).minusSeconds(1).toInstant().toEpochMilli();
            for (AgentAndTimeVO agentTime : agentTimeList) {
                if (agentTime.getStartTime() <= epochMilli) {
                    failRecords.add(newError(i + 1, "预约的考试时段，只能为" + allowCancelDays + "天之后的时段"));
                }
            }
        }
    }

    private TimePeriodEntity checkTimePeriod(String timePeriod, Map<String, TimePeriodEntity> timeCache) {
        if (timePeriod == null || timePeriod.split(" ").length != 2) {
            throw new StatusException(" 预约时段格式不正确");
        }
        String[] arr = timePeriod.split("-");
        String startTime = arr[0] + ":00";
        String endTime = startTime.substring(0, startTime.indexOf("日") + 1) + " " + arr[1] + ":00";
        Long startTimeLong = DateUtil.getLongTimeByZHDate(startTime);
        Long endTimeLong = DateUtil.getLongTimeByZHDate(endTime);
        if (timeCache.get(startTimeLong + "-" + endTimeLong) == null) {
            throw new StatusException(" 预约时段不存在");
        }
        return timeCache.get(startTimeLong + "-" + endTimeLong);
    }

    private Map<String, Object> newError(int lineNum, String msg) {
        Map<String, Object> map = new HashMap<>();
        map.put("lineNum", lineNum);
        map.put("msg", msg);
        return map;
    }

    private StudentEntity checkStudent(String studentCode, String studentName, String identityNumber, ApplyTaskEntity task) {
        LambdaQueryWrapper<StudentEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(StudentEntity::getStudentCode, studentCode);
        wrapper.eq(StudentEntity::getName, studentName);
        wrapper.eq(StudentEntity::getIdentityNumber, identityNumber);
        wrapper.eq(StudentEntity::getApplyTaskId, task.getId());
        return studentService.getOne(wrapper);
    }

    private String trimAndNullIfBlank(String s) {
        if (StringUtils.isBlank(s)) {
            return null;
        }
        return s.trim();
    }

    private Map<Long, Long> getExamSiteCategoryCache() {
        Map<Long, Long> cache = new HashMap<>();
        LambdaQueryWrapper<ExamSiteEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(ExamSiteEntity::getEnable, Boolean.TRUE);
        List<ExamSiteEntity> examSiteList = examSiteService.list(wrapper);
        examSiteList.forEach(site -> cache.put(site.getId(), site.getCategoryId()));
        return cache;
    }

    private Map<String, TimePeriodEntity> getTimePeriodCache() {
        Map<String, TimePeriodEntity> map = new HashMap<>();
        LambdaQueryWrapper<TimePeriodEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(TimePeriodEntity::getApplyTaskId, getApplyTask().getId());
        List<TimePeriodEntity> timeList = timePeriodService.list(wrapper);
        for (TimePeriodEntity time : timeList) {
            map.put(time.getStartTime() + "-" + time.getEndTime(), time);
        }
        return map;
    }

    private Map<String, Long> getAgentCache() {
        LambdaQueryWrapper<ExamSiteEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(ExamSiteEntity::getEnable, Boolean.TRUE);
        List<ExamSiteEntity> categoryList = examSiteService.list(wrapper);
        return categoryList.stream().collect(Collectors.toMap(ExamSiteEntity::getName, ExamSiteEntity::getId));
    }

    private Map<String, Long> getTeachingCache(Integer level) {
        LambdaQueryWrapper<CategoryEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(CategoryEntity::getEnable, Boolean.TRUE);
        wrapper.eq(CategoryEntity::getLevel, level == null ? CategoryLevel.TEACHING.getValue() : level);
        List<CategoryEntity> categoryList = categoryService.list(wrapper);
        return categoryList.stream().collect(Collectors.toMap(CategoryEntity::getName, CategoryEntity::getId));
    }

    private void checkOpenTime(Long orgId) {
        CurrentApplyTaskVO task = cacheService.currentApplyTask(orgId);
        Date startTime = new Date(task.getSelfApplyStartTime());
        Date endTime = new Date(task.getOpenApplyEndTime());
        Date now = new Date();
        if (now.before(startTime) || now.after(endTime)) {
            throw new StatusException("未到开放时间！");
        }
    }

    @Override
    public void autoAssign(Long taskId, Long userId) {
        checkAfterOpenTime();
        //未完成预约教学点
        List<CategoryVO> teachingList = studentService.listNoFinishCategory(taskId, Boolean.FALSE);
        List<TimePeriodEntity> timePeriodList = listTimePeriod(taskId);
        timePeriodList = listNoCancelTimePeriod(timePeriodList, taskId);
        if (timePeriodList == null || timePeriodList.isEmpty()) {
            log.warn("[autoAssign] 当前时间没有可预约的时段, taskId:{}, userId:{}", taskId, userId);
            throw new StatusException("当前时间没有可预约的时段");
        }

        //考位是否充足
        checkTeachingCapacity(taskId, teachingList, timePeriodList);

        RLock lock = (RLock) concurrentService.getLock(CacheConstants.LOCK_AUTO_APPLY);
        if (lock.isLocked()) {
            log.warn("[autoAssign] 获取锁失败, taskId:{}, lockKey:{}", taskId, CacheConstants.LOCK_AUTO_APPLY);
            throw new StatusException("自动分配执行未结束,请不要重复执行");
        }

        //异步执行
        studentAutoAssignService.autoAssign(taskId, userId, timePeriodList);
    }

    private void checkTeachingCapacity(Long taskId, List<CategoryVO> teachingList, List<TimePeriodEntity> timePeriodList) {
        for (CategoryVO categoryVO : teachingList) {
            List<ExamSiteEntity> examSiteList = listExamSite(categoryVO.getId(), null);
            if (examSiteList == null || examSiteList.isEmpty()) {
                log.warn("[autoAssign] teachingId: {} 下无考点", categoryVO.getId());
                continue;
            }

            //教学点所有的考位数量
            int total = examSiteList.stream().mapToInt(ExamSiteEntity::getCapacity).sum() * timePeriodList.size();
            List<Long> examSiteIds = examSiteList.stream().map(BaseEntity::getId).collect(Collectors.toList());
            List<Long> timePeriodIds = timePeriodList.stream().map(BaseEntity::getId).collect(Collectors.toList());
            //已预约的数量
            Integer haveApplyCount = baseMapper.getHaveApplyCount(examSiteIds, timePeriodIds, Boolean.FALSE);
            //未预约的数量
            Integer noApplyCount = baseMapper.getNoApplyCount(taskId, categoryVO.getId(), timePeriodIds);

            if (noApplyCount > total - haveApplyCount) {
                CategoryCacheBean categoryBean = categoryCacheService.getCategoryById(categoryVO.getId());
                log.warn("[autoAssign] {} 教学点剩余考位不足,剩余的考位数量:{},实际需要的考位数量:{}", categoryBean.getName(), total - haveApplyCount, noApplyCount);
                throw new StatusException("[" + categoryBean.getName() + "]教学点考位不足,剩余的考位数量:" + (total - haveApplyCount) + ",实际需要的考位数量:" + noApplyCount);
            }

        }
    }

    private List<ExamSiteEntity> listExamSite(Long teachingId, Long examSiteId) {
        LambdaQueryWrapper<ExamSiteEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(ExamSiteEntity::getCategoryId, teachingId);
        wrapper.eq(examSiteId != null, ExamSiteEntity::getId, examSiteId);
        wrapper.eq(ExamSiteEntity::getEnable, Boolean.TRUE);
        return examSiteService.list(wrapper);
    }

    private List<TimePeriodEntity> listNoCancelTimePeriod(List<TimePeriodEntity> timePeriodList, Long taskId) {
        ApplyTaskEntity task = applyTaskService.getById(taskId);
        Long longToday = DateUtil.getLongTimeByDate(DateUtil.formatShortSplitDateString(new Date()) + " 00:00:00");
        Date today = new Date(longToday);
        Date otherDay = DateUtil.addValues(today, Calendar.DAY_OF_MONTH, task.getAllowApplyCancelDays());
        Long longOtherDay = DateUtil.getLongTimeByDate(DateUtil.formatShortSplitDateString(otherDay) + " 23:59:59");
        return timePeriodList.stream().filter(time -> time.getStartTime() > longOtherDay).collect(Collectors.toList());
    }

    private List<TimePeriodEntity> listTimePeriod(Long taskId) {
        LambdaQueryWrapper<TimePeriodEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(TimePeriodEntity::getApplyTaskId, taskId);
        wrapper.orderByAsc(TimePeriodEntity::getStartTime);
        List<TimePeriodEntity> timeList = timePeriodService.list(wrapper);
        if (timeList.isEmpty()) {
            log.warn("当前任务未配置可用的考试时段,taskId:{}", taskId);
            throw new StatusException("考试时段未设置");
        }
        return timeList;
    }

    private void checkAfterOpenTime() {
        ApplyTaskEntity task = getApplyTask();
        Date selfEndTime = new Date(task.getSelfApplyEndTime());
        Date openStartTime = new Date(task.getOpenApplyStartTime());
        if (!DateUtil.isBetwwen(selfEndTime, openStartTime)) {
            throw new StatusException("自动分配，时间必须要在第一阶段结束之后，第三阶段开始之前");
        }
    }

    @Override
    public void autoLayout(Long teachingId) {

    }

    @Override
    public File downloadSignIn(Long teachingId, Long agentId, Long examDate) {
        ApplyTaskEntity task = getApplyTask();
        List<TimePeriodEntity> timePeriodList = listTimePeriod(task.getId());
        List<TimePeriodEntity> sameDayTimePeriodList = listSameDayTimePeriod(timePeriodList, examDate);
        if (sameDayTimePeriodList.isEmpty()) {
            log.warn("[downloadSignIn] 没有查询到当天的时段,examDate:{}", examDate);
            throw new StatusException("没有可下载的签到表");
        }
        List<ExamSiteEntity> examSiteList = listExamSite(teachingId, agentId);
        if (examSiteList == null || examSiteList.isEmpty()) {
            log.warn("[downloadSignIn] teachingId:{},agentId:{} 没有可用的考点", teachingId, agentId);
            throw new StatusException("没有可用的考点");
        }
        File tempDir = new File("temp");
        if (!tempDir.exists()) {
            try {
                boolean success = tempDir.mkdirs();
                if (!success) {
                    log.warn("[downloadSignIn] 临时目录创建失败!");
                }
            } catch (Exception e) {
                log.error("[downloadSignIn] 临时目录创建失败,msg:{}", e.getMessage());
            }
        }

        CategoryCacheBean category = categoryCacheService.getCategoryById(teachingId);
        File zipFile = new File(tempDir, category.getName() + "签到表.zip");
        ZipWriter writer;
        try {
            writer = ZipWriter.create(zipFile);
            List<File> fileList = new ArrayList<>();
            for (ExamSiteEntity examSite : examSiteList) {
                fileList.addAll(downloadByExamSite(writer, task.getName(), examSite, sameDayTimePeriodList));
            }
            if (fileList.isEmpty()) {
                throw new StatusException("暂未排考,请等待排考后下载");
            }
        } catch (IOException e) {
            log.error("[downloadSignIn] 创建文件异常,msg:{}", e.getMessage());
            throw new StatusException("文件写入异常");
        }
        writer.close();
        return zipFile;
    }

    public List<File> downloadByExamSite(ZipWriter writer, String taskName, ExamSiteEntity site, List<TimePeriodEntity> timeList) throws IOException {
        List<File> fileList = new ArrayList<>();
        List<ExamRoomEntity> roomList = listExamRoom(site.getId());
        MaterialTitleInfo title = new MaterialTitleInfo();
        title.setTaskName(taskName);
        title.setSiteName(site.getName());
        for (TimePeriodEntity time : timeList) {
            title.setTimePeriod(DateUtil.getStartAndEndTime(time.getStartTime(), time.getEndTime()));
            for (ExamRoomEntity room : roomList) {
                title.setAddress(room.getAddress());
                title.setRoomCode(room.getCode());
                title.setRoomName(room.getName());
                List<StudentApplyVO> studentList = baseMapper.listStudentApply(time.getId(), room.getId());
                if (!studentList.isEmpty()) {
                    File file = materialGenerateService.generateSignInForm(title, studentList);
                    fileList.add(file);
                    writer.write(file, title.getSiteName(), title.getTimePeriod() + " 第【" + title.getRoomCode() + "】考场" + ".pdf");
                }
            }
        }
        return fileList;
    }

    public List<ExamRoomEntity> listExamRoom(Long examSiteId) {
        LambdaQueryWrapper<ExamRoomEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(ExamRoomEntity::getExamSiteId, examSiteId);
        wrapper.eq(ExamRoomEntity::getEnable, Boolean.TRUE);
        wrapper.orderByAsc(ExamRoomEntity::getCode);
        return examRoomService.list(wrapper);
    }

    private List<TimePeriodEntity> listSameDayTimePeriod(List<TimePeriodEntity> timePeriodList, Long startTime) {
        Date day = new Date(startTime);
        List<TimePeriodEntity> resultList = new ArrayList<>();
        for (TimePeriodEntity time : timePeriodList) {
            if (isSameDay(day, new Date(time.getStartTime())))
                resultList.add(time);
        }
        return resultList.stream().sorted(Comparator.comparing(TimePeriodEntity::getStartTime)).collect(Collectors.toList());
    }

    @Override
    public int countApplyFinishForExamSiteAndTimePeriod(Long examSiteId, Long timePeriodId) {
        Integer value = baseMapper.countApplyFinishForExamSiteAndTimePeriod(examSiteId, timePeriodId);
        return value != null ? value : 0;
    }

    @Override
    public List<SignInVO> listSignInDate(Long taskId) {
        ApplyTaskEntity task;
        if (taskId != null) {
            task = applyTaskService.getById(taskId);
        } else {
            LambdaQueryWrapper<ApplyTaskEntity> wrapper = new LambdaQueryWrapper<>();
            wrapper.eq(ApplyTaskEntity::getEnable, Boolean.TRUE);
            task = applyTaskService.getOne(wrapper);
        }

        if (task == null) {
            log.warn("[listSignDate] 未开启预约任务");
            return Collections.emptyList();
        }
        List<TimePeriodEntity> timePeriodList = listTimePeriod(task.getId());
        if (timePeriodList.isEmpty()) {
            log.warn("[listSignDate] 未配置考试时段");
            return Collections.emptyList();
        }

        List<SignInVO> signInVoList = new ArrayList<>();
        Date today = DateUtils.truncate(new Date(), Calendar.DATE);
        if (isInTimePeriod(today, timePeriodList)) {
            SignInVO signInVO = new SignInVO();
            signInVO.setExamDate(today.getTime());
            signInVoList.add(signInVO);
        }
        for (int i = 1; i <= task.getAllowApplyCancelDays(); i++) {
            Date otherDay = DateUtils.addDays(today, i);
            if (isInTimePeriod(otherDay, timePeriodList)) {
                SignInVO signInVO = new SignInVO();
                signInVO.setExamDate(otherDay.getTime());
                signInVoList.add(signInVO);
            }
        }

        signInVoList.sort(Comparator.comparing(SignInVO::getExamDate));
        return signInVoList;
    }

    private boolean isInTimePeriod(Date date, List<TimePeriodEntity> timePeriodList) {
        for (TimePeriodEntity timePeriod : timePeriodList) {
            Date day = new Date(timePeriod.getStartTime());
            if (DateUtils.isSameDay(day, date)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public List<String[]> listExamSiteAvailable(List<CategoryVO> examSiteList) {
        List<String[]> availableList = new ArrayList<>();
        ApplyTaskEntity task = getApplyTask();
        List<TimePeriodEntity> timePeriodList = listTimePeriod(task.getId());
        for (TimePeriodEntity timePeriod : timePeriodList) {
            String[] availableArr = getAvailableArr(examSiteList, timePeriod);
            availableList.add(availableArr);
        }
        return availableList;
    }

    private String[] getAvailableArr(List<CategoryVO> examSiteList, TimePeriodEntity timePeriod) {
        String[] availableArr = new String[examSiteList.size() + 1];
        availableArr[0] = DateUtil.getStartAndEndTime(timePeriod.getStartTime(), timePeriod.getEndTime());
        for (int i = 0; i < examSiteList.size(); i++) {
            CategoryVO categoryVO = examSiteList.get(i);
            int applyFinishCount = countApplyFinishForExamSiteAndTimePeriod(categoryVO.getId(), timePeriod.getId());
            //剩余的容量
            availableArr[i + 1] = String.valueOf(categoryVO.getCapacity() - applyFinishCount);
        }
        return availableArr;
    }

    @Override
    public List<StudentApplyExport> listStudentApplyExport(Long teachingId) {
        return Collections.emptyList();
    }

    @Override
    public List<StudentExport> listStudentNoApply(Long teachingId) {
        return Collections.emptyList();
    }

    @Override
    public void exportNoApplyStudent(Long teachingId, Long operateId) {
        //写入异步任务
        Map<String, Object> taskMap = asyncTaskService.saveAsyncTask(AsyncTaskType.NO_FINISH_APPLY_STUDENT,
                operateId);
        taskMap.computeIfAbsent("teachingId", v -> teachingId);
        taskMap.computeIfAbsent("operateId", v -> operateId);
        try {
            studentApplyNoFinishExportService.exportTask(taskMap);
        } catch (Exception e) {
            throw new StatusException(e.getMessage());
        }
    }
}
